/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/02/05 16:12:14
 */

import { DOMParser, XMLSerializer } from '@xmldom/xmldom'
import svgpath from 'svgpath'

const shape = {
  $translatePath(convertObj, attributes) {
    let data = {}
    attributes.forEach(function (attribute) {
      data[attribute.name] = attribute.value
    })

    return convertObj.template(data)
  },
  generatePathNode(doc, node) {
    let convertObj, attributes, path, tagName

    tagName = String(node.tagName).toLowerCase()
    //polyline与polygon是同一个东西
    if (tagName === 'polyline') {
      tagName = 'polygon'
    }

    convertObj = shape[tagName]
    if (!convertObj) {
      return
    }

    path = shape.$translatePath(convertObj, node.attributes)
    if (!path) {
      return
    }

    newPathNode = doc.createElement('path')
    //赋值d属性
    newPathNode.setAttribute('d', path)
    //保留其他属性
    node.attributes.forEach(function (attribute) {
      if (!convertObj.attrs.includes(attribute.name)) {
        newPathNode.setAttribute(attribute.name, attribute.value)
      }
    })

    return newPathNode
  },
  rect: {
    attrs: ['x', 'y', 'width', 'height', 'rx', 'ry'],
    template(data) {
      // 需要做一个保护，如果没有写属性值那么默认为0
      if (!data.x) {
        data.x = '0'
      }

      if (!data.y) {
        data.y = '0'
      }

      if (!data.rx && data.ry) {
        data.rx = data.ry
      }

      if (!data.ry && data.rx) {
        data.ry = data.rx
      }

      if (!data.ry && !data.rx) {
        data.ry = data.rx = '0'
      }

      return `M${data.x},${data.y}m${data.rx},0l${data.width - 2 * data.rx},0q${
        data.rx
      },0 ${data.rx},${data.ry}l0,${data.height - 2 * data.ry}q0,${
        data.ry
      } ${-data.rx},${data.ry}l${
        -data.width + 2 * data.rx
      },0q${-data.rx},0 ${-data.rx},${-data.ry}l0,${
        -data.height + 2 * data.ry
      }q0,${-data.ry},${data.rx},${-data.ry}Z`
    }
  },
  circle: {
    attrs: ['cx', 'cy', 'r'],
    template(data) {
      return `M${data.cx},${data.cy}m${-data.r},0a${data.r},${data.r},0,1,0,${
        data.r * 2
      },0a${data.r},${data.r},0,1,0,-${data.r * 2},0Z`
    }
  },
  line: {
    attrs: ['x1', 'y1', 'x2', 'y2'],
    template(data) {
      return `M${data.x1},${data.y1}L${data.x2},${data.y2}L${data.x1},${data.y1}Z`
    }
  },
  ellipse: {
    attrs: ['cx', 'cy', 'rx', 'ry'],
    template(data) {
      return `M${data.cx - data.rx},${data.cy}a${data.rx},${data.ry},0,1,0,${
        data.rx * 2
      },0a${data.rx},${data.ry},0,1,0,-${data.rx * 2},0Z`
    }
  },
  polygon: {
    attrs: ['points'],
    template(data) {
      data.points = data.points.replace(/(^\s*)|(\s*$)/g, '')
      let points = data.points.split(' ')
      let p = ''
      points.forEach(function (point, index) {
        if (index === 0) {
          p += 'M' + point
        } else {
          p += 'L' + point
        }
      })
      p += 'Z'

      return p
    }
  }
}

function generatePath(path, transforms) {
  let t = svgpath(path).abs()
  transforms.forEach(it => t.transform(it))
  return t.toString()
}

export function normalizePath(node, transforms) {
  let path = ''
  let newTransForms = transforms ? [...transforms] : []

  if (node.getAttribute && node.getAttribute('transform')) {
    newTransForms.push(node.getAttribute('transform'))
  }

  if (!node.hasChildNodes() && node.tagName === 'path') {
    path = generatePath(node.getAttribute('d'), newTransForms)
  }

  if (node.hasChildNodes()) {
    node.childNodes.forEach(function (childNode) {
      path += normalizePath(childNode, newTransForms)
    })
  }

  return path
}

function transNode(doc, node) {
  let convertObj, newPathNode

  //基本的图形都是没有子结点的
  if (!node.hasChildNodes() && node.nodeName !== 'path') {
    newPathNode = shape.generatePathNode(doc, node)
    newPathNode && node.parentNode.replaceChild(newPathNode, node)
    return
  }

  node.childNodes.forEach(function (child) {
    transNode(doc, child)
  })
}

export function pathify(svgString) {
  let doc = new DOMParser().parseFromString(svgString, 'application/xml')

  transNode(doc, doc)

  return new XMLSerializer().serializeToString(doc)
}
