// Light implementation of binary buffer with helpers for easy access.
// https://github.com/fontello/microbuffer

function createArray(size) {
  return new Uint8Array(size)
}

export class ByteBuffer {
  constructor(buffer, start, length) {
    let isInherited = buffer instanceof MicroBuffer

    this.buffer = isInherited
      ? buffer.buffer
      : typeof buffer === 'number'
      ? createArray(buffer)
      : buffer

    this.start = (start || 0) + (isInherited ? buffer.start : 0)
    this.length = length || this.buffer.length - this.start
    this.offset = 0

    this.isTyped = !Array.isArray(this.buffer)
  }
  getUint8(pos) {
    return this.buffer[pos + this.start]
  }

  getUint16(pos, littleEndian) {
    let val
    if (littleEndian) {
      throw new Error('not implemented')
    } else {
      val = this.buffer[pos + 1 + this.start]
      val += (this.buffer[pos + this.start] << 8) >>> 0
    }
    return val
  }

  getUint32(pos, littleEndian) {
    let val
    if (littleEndian) {
      throw new Error('not implemented')
    } else {
      val = this.buffer[pos + 1 + this.start] << 16
      val |= this.buffer[pos + 2 + this.start] << 8
      val |= this.buffer[pos + 3 + this.start]
      val += (this.buffer[pos + this.start] << 24) >>> 0
    }
    return val
  }

  setUint8(pos, value) {
    this.buffer[pos + this.start] = value & 0xff
  }

  setUint16(pos, value, littleEndian) {
    let offset = pos + this.start
    let buf = this.buffer
    if (littleEndian) {
      buf[offset] = value & 0xff
      buf[offset + 1] = (value >>> 8) & 0xff
    } else {
      buf[offset] = (value >>> 8) & 0xff
      buf[offset + 1] = value & 0xff
    }
  }

  setUint32(pos, value, littleEndian) {
    let offset = pos + this.start
    let buf = this.buffer
    if (littleEndian) {
      buf[offset] = value & 0xff
      buf[offset + 1] = (value >>> 8) & 0xff
      buf[offset + 2] = (value >>> 16) & 0xff
      buf[offset + 3] = (value >>> 24) & 0xff
    } else {
      buf[offset] = (value >>> 24) & 0xff
      buf[offset + 1] = (value >>> 16) & 0xff
      buf[offset + 2] = (value >>> 8) & 0xff
      buf[offset + 3] = value & 0xff
    }
  }

  writeUint8(value) {
    this.buffer[this.offset + this.start] = value & 0xff
    this.offset++
  }

  writeInt8(value) {
    this.setUint8(this.offset, value < 0 ? 0xff + value + 1 : value)
    this.offset++
  }

  writeUint16(value, littleEndian) {
    this.setUint16(this.offset, value, littleEndian)
    this.offset += 2
  }

  writeInt16(value, littleEndian) {
    this.setUint16(
      this.offset,
      value < 0 ? 0xffff + value + 1 : value,
      littleEndian
    )
    this.offset += 2
  }

  writeUint32(value, littleEndian) {
    this.setUint32(this.offset, value, littleEndian)
    this.offset += 4
  }

  writeInt32(value, littleEndian) {
    this.setUint32(
      this.offset,
      value < 0 ? 0xffffffff + value + 1 : value,
      littleEndian
    )
    this.offset += 4
  }

  // get current position
  //
  tell() {
    return this.offset
  }

  // set current position
  //
  seek(pos) {
    this.offset = pos
  }

  fill(value) {
    let index = this.length - 1
    while (index >= 0) {
      this.buffer[index + this.start] = value
      index--
    }
  }

  writeUint64(value) {
    // we canot use bitwise operations for 64bit values because of JavaScript limitations,
    // instead we should divide it to 2 Int32 numbers
    // 2^32 = 4294967296
    let hi = Math.floor(value / 4294967296)
    let lo = value - hi * 4294967296
    this.writeUint32(hi)
    this.writeUint32(lo)
  }

  writeBytes(data) {
    let buffer = this.buffer
    let offset = this.offset + this.start
    if (this.isTyped) {
      buffer.set(data, offset)
    } else {
      for (let i = 0; i < data.length; i++) {
        buffer[i + offset] = data[i]
      }
    }
    this.offset += data.length
  }

  toString(offset, length) {
    // default values if not set
    offset = offset || 0
    length = length || this.length - offset

    // add buffer shift
    let start = offset + this.start
    let end = start + length

    let string = ''
    for (let i = start; i < end; i++) {
      string += String.fromCharCode(this.buffer[i])
    }
    return string
  }

  toArray() {
    if (this.isTyped) {
      return this.buffer.subarray(this.start, this.start + this.length)
    }

    return this.buffer.slice(this.start, this.start + this.length)
  }
}
