/**
 * {description of this file}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/02/04 12:16:26
 */

export class Font {
  ascent = 850
  copyright = ''
  createdDate = new Date()
  glyphs = []
  ligatures = []
  // Maping of code points to glyphs.
  // Keys are actually numeric, thus should be `parseInt`ed.
  codePoints = {}
  isFixedPitch = 0
  italicAngle = 0
  familyClass = 0 // No Classification
  familyName = ''

  // 0x40 - REGULAR - Characters are in the standard weight/style for the font
  // 0x80 - USE_TYPO_METRICS - use OS/2.sTypoAscender - OS/2.sTypoDescender + OS/2.sTypoLineGap as the default line spacing
  // https://docs.microsoft.com/en-us/typography/opentype/spec/os2#fsselection
  // https://github.com/fontello/svg2ttf/issues/95
  fsSelection = 0x40 | 0x80

  // Non zero value can cause issues in IE, https://github.com/fontello/svg2ttf/issues/45
  fsType = 0
  lowestRecPPEM = 8
  macStyle = 0
  modifiedDate = new Date()
  panose = {
    familyType: 2, // Latin Text
    serifStyle: 0, // any
    weight: 5, // book
    proportion: 3, //modern
    contrast: 0, //any
    strokeVariation: 0, //any,
    armStyle: 0, //any,
    letterform: 0, //any,
    midline: 0, //any,
    xHeight: 0 //any,
  }
  revision = 1
  sfntNames = []
  underlineThickness = 0
  unitsPerEm = 1000
  weightClass = 400 // normal
  width = 1000
  widthClass = 5 // Medium (normal)
  ySubscriptXOffset = 0
  ySuperscriptXOffset = 0
  int_descent = -150
  xHeight = 0
  capHeight = 0

  /*  */

  get descent() {
    return this.int_descent
  }

  set descent(value) {
    this.int_descent = parseInt(Math.round(-Math.abs(value)), 10)
  }

  get avgCharWidth() {
    if (this.glyphs.length === 0) {
      return 0
    }
    let widths = this.glyphs.map(it => it.width)

    return ~~(
      widths.reduce(function (prev, cur) {
        return prev + cur
      }) / widths.length
    )
  }

  get ySubscriptXSize() {
    return ~~(void 0 === this.int_ySubscriptXSize
      ? this.width * 0.6347
      : this.int_ySubscriptXSize)
  }
  set ySubscriptXSize(value) {
    this.int_ySubscriptXSize = value
  }

  get ySubscriptYSize() {
    return ~~(void 0 === this.int_ySubscriptYSize
      ? (this.ascent - this.descent) * 0.7
      : this.int_ySubscriptYSize)
  }
  set ySubscriptYSize(value) {
    this.int_ySubscriptYSize = value
  }

  get ySubscriptYOffset() {
    return ~~(void 0 === this.int_ySubscriptYOffset
      ? (this.ascent - this.descent) * 0.14
      : this.int_ySubscriptYOffset)
  }
  set ySubscriptYOffset(value) {
    this.int_ySubscriptYOffset = value
  }

  get ySuperscriptXSize() {
    return ~~(void 0 === this.int_ySuperscriptXSize
      ? this.width * 0.6347
      : this.int_ySuperscriptXSize)
  }
  set ySuperscriptXSize(value) {
    this.int_ySuperscriptXSize = value
  }

  get ySuperscriptYSize() {
    return ~~(void 0 === this.int_ySuperscriptYSize
      ? (this.ascent - this.descent) * 0.7
      : this.int_ySuperscriptYSize)
  }
  set ySuperscriptYSize(value) {
    this.int_ySuperscriptYSize = value
  }

  get ySuperscriptYOffset() {
    return ~~(void 0 === this.int_ySuperscriptYOffset
      ? (this.ascent - this.descent) * 0.48
      : this.int_ySuperscriptYOffset)
  }
  set ySuperscriptYOffset(value) {
    this.int_ySuperscriptYOffset = value
  }

  get yStrikeoutSize() {
    return ~~(void 0 === this.int_yStrikeoutSize
      ? (this.ascent - this.descent) * 0.049
      : this.int_yStrikeoutSize)
  }
  set yStrikeoutSize(value) {
    this.int_yStrikeoutSize = value
  }

  get yStrikeoutPosition() {
    return ~~(void 0 === this.int_yStrikeoutPosition
      ? (this.ascent - this.descent) * 0.258
      : this.int_yStrikeoutPosition)
  }
  set yStrikeoutPosition(value) {
    this.int_yStrikeoutPosition = value
  }

  get minLsb() {
    return ~~Math.min(...this.glyphs.map(it => it.xMin))
  }

  get minRsb() {
    if (!this.glyphs.length) return ~~this.width

    return ~~this.glyphs.reduce(function (minRsb, glyph) {
      return Math.min(minRsb, glyph.width - glyph.xMax)
    }, 0)
  }

  get xMin() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (xMin, glyph) {
      return Math.min(xMin, glyph.xMin)
    }, 0)
  }

  get yMin() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (yMin, glyph) {
      return Math.min(yMin, glyph.yMin)
    }, 0)
  }

  get xMax() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (xMax, glyph) {
      return Math.max(xMax, glyph.xMax)
    }, 0)
  }

  get yMax() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (yMax, glyph) {
      return Math.max(yMax, glyph.yMax)
    }, 0)
  }

  get avgWidth() {
    let len = this.glyphs.length

    if (len === 0) {
      return this.width
    }

    let sumWidth = this.glyphs.reduce(function (sumWidth, glyph) {
      return sumWidth + glyph.width
    }, 0)

    return Math.round(sumWidth / len)
  }

  get maxWidth() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (maxWidth, glyph) {
      return Math.max(maxWidth, glyph.width)
    }, 0)
  }

  get maxExtent() {
    if (!this.glyphs.length) return this.width

    return this.glyphs.reduce(function (maxExtent, glyph) {
      return Math.max(maxExtent, glyph.xMax /*- glyph.xMin*/)
    }, 0)
  }

  // Property used for `sTypoLineGap` in OS/2 and not used for `lineGap` in HHEA, because
  // non zero lineGap causes bad offset in IE, https://github.com/fontello/svg2ttf/issues/37

  get lineGap() {
    return ~~(void 0 === this.int_lineGap
      ? (this.ascent - this.descent) * 0.09
      : this.int_lineGap)
  }
  set lineGap(value) {
    this.int_lineGap = value
  }

  get underlinePosition() {
    return ~~(void 0 === this.int_underlinePosition
      ? (this.ascent - this.descent) * 0.01
      : this.int_underlinePosition)
  }
  set underlinePosition(value) {
    this.int_underlinePosition = value
  }
}

export function Glyph() {
  this.contours = []
  this.d = ''
  this.id = ''
  this.codes = [] // needed for nice validator error output
  this.height = 0
  this.name = ''
  this.width = 0
}

Object.defineProperty(Glyph.prototype, 'xMin', {
  get: function () {
    let xMin = 0
    let hasPoints = false

    this.contours.forEach(function (contour) {
      contour.points.forEach(function (point) {
        xMin = Math.min(xMin, Math.floor(point.x))
        hasPoints = true
      })
    })

    if (xMin < -32768) {
      throw new Error(
        'xMin value for glyph ' +
          (this.name ? '"' + this.name + '"' : JSON.stringify(this.codes)) +
          ' is out of bounds (actual ' +
          xMin +
          ', expected -32768..32767, d="' +
          this.d +
          '")'
      )
    }
    return hasPoints ? xMin : 0
  }
})

Object.defineProperty(Glyph.prototype, 'xMax', {
  get: function () {
    let xMax = 0
    let hasPoints = false

    this.contours.forEach(function (contour) {
      contour.points.forEach(function (point) {
        xMax = Math.max(xMax, -Math.floor(-point.x))
        hasPoints = true
      })
    })

    if (xMax > 32767) {
      throw new Error(
        'xMax value for glyph ' +
          (this.name ? '"' + this.name + '"' : JSON.stringify(this.codes)) +
          ' is out of bounds (actual ' +
          xMax +
          ', expected -32768..32767, d="' +
          this.d +
          '")'
      )
    }
    return hasPoints ? xMax : this.width
  }
})

Object.defineProperty(Glyph.prototype, 'yMin', {
  get: function () {
    let yMin = 0
    let hasPoints = false

    this.contours.forEach(function (contour) {
      contour.points.forEach(function (point) {
        yMin = Math.min(yMin, Math.floor(point.y))
        hasPoints = true
      })
    })

    if (yMin < -32768) {
      throw new Error(
        'yMin value for glyph ' +
          (this.name ? '"' + this.name + '"' : JSON.stringify(this.codes)) +
          ' is out of bounds (actual ' +
          yMin +
          ', expected -32768..32767, d="' +
          this.d +
          '")'
      )
    }
    return hasPoints ? yMin : 0
  }
})

Object.defineProperty(Glyph.prototype, 'yMax', {
  get: function () {
    let yMax = 0
    let hasPoints = false

    this.contours.forEach(function (contour) {
      contour.points.forEach(function (point) {
        yMax = Math.max(yMax, -Math.floor(-point.y))
        hasPoints = true
      })
    })

    if (yMax > 32767) {
      throw new Error(
        'yMax value for glyph ' +
          (this.name ? '"' + this.name + '"' : JSON.stringify(this.codes)) +
          ' is out of bounds (actual ' +
          yMax +
          ', expected -32768..32767, d="' +
          this.d +
          '")'
      )
    }
    return hasPoints ? yMax : 0
  }
})

export function Contour() {
  this.points = []
}

export function Point() {
  this.onCurve = true
  this.x = 0
  this.y = 0
}

export function SfntName() {
  this.id = 0
  this.value = ''
}
