/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/02/04 11:34:19
 */

import { ByteBuffer } from './microbuffer.js'

import createGSUBTable from './ttf/tables/gsub.js'
import createOS2Table from './ttf/tables/os2.js'
import createCMapTable from './ttf/tables/cmap.js'
import createGlyfTable from './ttf/tables/glyf.js'
import createHeadTable from './ttf/tables/head.js'
import createHHeadTable from './ttf/tables/hhea.js'
import createHtmxTable from './ttf/tables/hmtx.js'
import createLocaTable from './ttf/tables/loca.js'
import createMaxpTable from './ttf/tables/maxp.js'
import createNameTable from './ttf/tables/name.js'
import createPostTable from './ttf/tables/post.js'

import * as utils from './ttf/utils.js'

// Tables
let TABLES = [
  { innerName: 'hhea', order: 1, create: createHHeadTable }, // hhea
  { innerName: 'head', order: 2, create: createHeadTable }, // head
  { innerName: 'maxp', order: 3, create: createMaxpTable }, // maxp
  { innerName: 'GSUB', order: 4, create: createGSUBTable }, // GSUB
  { innerName: 'OS/2', order: 4, create: createOS2Table }, // OS/2
  { innerName: 'hmtx', order: 5, create: createHtmxTable }, // hmtx
  { innerName: 'cmap', order: 6, create: createCMapTable }, // cmap
  { innerName: 'loca', order: 7, create: createLocaTable }, // loca
  { innerName: 'glyf', order: 8, create: createGlyfTable }, // glyf
  { innerName: 'name', order: 9, create: createNameTable }, // name
  { innerName: 'post', order: 10, create: createPostTable } // post
]

// Various constants
let CONST = {
  VERSION: 0x10000,
  CHECKSUM_ADJUSTMENT: 0xb1b0afba
}

function ulong(t) {
  t &= 0xffffffff
  if (t < 0) {
    t += 0x100000000
  }
  return t
}

function calc_checksum(buf) {
  let sum = 0
  let nlongs = Math.floor(buf.length / 4)

  for (let i = 0; i < nlongs; ++i) {
    let t = buf.getUint32(i * 4)

    sum = ulong(sum + t)
  }

  let leftBytes = buf.length - nlongs * 4 //extra 1..3 bytes found, because table is not aligned. Need to include them in checksum too.

  if (leftBytes > 0) {
    let leftRes = 0

    for (let i = 0; i < 4; i++) {
      leftRes =
        (leftRes << 8) + (i < leftBytes ? buf.getUint8(nlongs * 4 + i) : 0)
    }
    sum = ulong(sum + leftRes)
  }
  return sum
}

export function generateTTF(font) {
  // Prepare TTF contours objects. Note, that while sfnt countours are classes,
  // ttf contours are just plain arrays of points
  font.glyphs.forEach(function (glyph) {
    glyph.ttfContours = glyph.contours.map(function (contour) {
      return contour.points
    })
    // 0.3px accuracy is ok. fo 1000x1000.
    glyph.ttfContours = utils.simplify(glyph.ttfContours, 0.3)
    glyph.ttfContours = utils.simplify(glyph.ttfContours, 0.3) // one pass is not enougth

    // Interpolated points can be removed. 1.1px is acceptable
    // measure - it will give us 1px error after coordinates rounding.
    glyph.ttfContours = utils.interpolate(glyph.ttfContours, 1.1)

    glyph.ttfContours = utils.roundPoints(glyph.ttfContours)
    glyph.ttfContours = utils.removeClosingReturnPoints(glyph.ttfContours)
    glyph.ttfContours = utils.toRelative(glyph.ttfContours)
  })

  // Add tables
  let headerSize = 12 + 16 * TABLES.length // TTF header plus table headers
  let bufSize = headerSize
  //calculate offsets
  let offset = headerSize

  TABLES.forEach(function (table) {
    //store each table in its own buffer
    table.buffer = table.create(font)
    table.length = table.buffer.length
    table.corLength = table.length + ((4 - (table.length % 4)) % 4) // table size should be divisible to 4
    table.checkSum = calc_checksum(table.buffer)
    bufSize += table.corLength

    table.offset = offset
    offset += table.corLength
  })

  //create TTF buffer

  let buf = new ByteBuffer(bufSize)

  //special constants
  let entrySelector = Math.floor(Math.log(TABLES.length) / Math.LN2)
  let searchRange = Math.pow(2, entrySelector) * 16
  let rangeShift = TABLES.length * 16 - searchRange

  // Add TTF header
  buf.writeUint32(CONST.VERSION)
  buf.writeUint16(TABLES.length)
  buf.writeUint16(searchRange)
  buf.writeUint16(entrySelector)
  buf.writeUint16(rangeShift)

  let headOffset = 0

  TABLES.forEach(function (table) {
    buf.writeUint32(utils.identifier(table.innerName)) //inner name
    buf.writeUint32(table.checkSum) //checksum
    buf.writeUint32(table.offset) //offset
    buf.writeUint32(table.length) //length

    if (table.innerName === 'head') {
      //we must store head offset to write font checksum
      headOffset = buf.tell()
    }
    buf.writeBytes(table.buffer.buffer)
    for (let i = table.length; i < table.corLength; i++) {
      //align table to be divisible to 4
      buf.writeUint8(0)
    }
  })

  // Write font checksum (corrected by magic value) into HEAD table
  buf.setUint32(
    headOffset + 8,
    ulong(CONST.CHECKSUM_ADJUSTMENT - calc_checksum(buf))
  )

  return buf
}
