// See documentation here: http://www.microsoft.com/typography/otspec/GSUB.htm

import { identifier } from '../utils.js'
import { ByteBuffer } from '../../microbuffer.js'

function createScript() {
  let scriptRecord =
    0 +
    2 + // Script DefaultLangSys Offset
    2 // Script[0] LangSysCount (0)

  let langSys =
    0 +
    2 + // Script DefaultLangSys LookupOrder
    2 + // Script DefaultLangSys ReqFeatureIndex
    2 + // Script DefaultLangSys FeatureCount (0?)
    2 // Script Optional Feature Index[0]

  let length = 0 + scriptRecord + langSys

  let buffer = new ByteBuffer(length)

  // Script Record
  // Offset to the start of langSys from the start of scriptRecord
  buffer.writeUint16(scriptRecord) // DefaultLangSys

  // Number of LangSys entries other than the default (none)
  buffer.writeUint16(0)

  // LangSys record (DefaultLangSys)
  // LookupOrder
  buffer.writeUint16(0)
  // ReqFeatureIndex -> only one required feature: all ligatures
  buffer.writeUint16(0)
  // Number of FeatureIndex values for this language system (excludes the required feature)
  buffer.writeUint16(1)
  // FeatureIndex for the first optional feature
  // Note: Adding the same feature to both the optional
  // and the required features is a clear violation of the spec
  // but it fixes IE not displaying the ligatures.
  // See http://partners.adobe.com/public/developer/opentype/index_table_formats.html, Section “Language System Table”
  // “FeatureCount: Number of FeatureIndex values for this language system-*excludes the required feature*” (emphasis added)
  buffer.writeUint16(0)

  return buffer
}

function createScriptList() {
  let scriptSize =
    0 +
    4 + // Tag
    2 // Offset

  // tags should be arranged alphabetically
  let scripts = [
    ['DFLT', createScript()],
    ['latn', createScript()]
  ]

  let header =
    0 +
    2 + // Script count
    scripts.length * scriptSize

  let tableLengths = scripts
    .map(function (script) {
      return script[1].length
    })
    .reduce(function (result, count) {
      return result + count
    }, 0)

  let length = 0 + header + tableLengths

  let buffer = new ByteBuffer(length)

  // Script count
  buffer.writeUint16(scripts.length)

  // Write all ScriptRecords
  let offset = header

  scripts.forEach(function (script) {
    let name = script[0],
      table = script[1]

    // Script identifier (DFLT/latn)
    buffer.writeUint32(identifier(name))
    // Offset to the ScriptRecord from start of the script list
    buffer.writeUint16(offset)
    // Increment offset by script table length
    offset += table.length
  })

  // Write all ScriptTables
  scripts.forEach(function (script) {
    let table = script[1]

    buffer.writeBytes(table.buffer)
  })

  return buffer
}

// Write one feature containing all ligatures
function createFeatureList() {
  let header =
    0 +
    2 + // FeatureCount
    4 + // FeatureTag[0]
    2 // Feature Offset[0]

  let length =
    0 +
    header +
    2 + // FeatureParams[0]
    2 + // LookupCount[0]
    2 // Lookup[0] LookupListIndex[0]

  let buffer = new ByteBuffer(length)

  // FeatureCount
  buffer.writeUint16(1)
  // FeatureTag[0]
  buffer.writeUint32(identifier('liga'))
  // Feature Offset[0]
  buffer.writeUint16(header)
  // FeatureParams[0]
  buffer.writeUint16(0)
  // LookupCount[0]
  buffer.writeUint16(1)
  // Index into lookup table. Since we only have ligatures, the index is always 0
  buffer.writeUint16(0)

  return buffer
}

function createLigatureCoverage(font, ligatureGroups) {
  let glyphCount = ligatureGroups.length

  let length =
    0 +
    2 + // CoverageFormat
    2 + // GlyphCount
    2 * glyphCount // GlyphID[i]

  let buffer = new ByteBuffer(length)

  // CoverageFormat
  buffer.writeUint16(1)

  // Length
  buffer.writeUint16(glyphCount)

  ligatureGroups.forEach(function (group) {
    buffer.writeUint16(group.startGlyph.id)
  })

  return buffer
}

function createLigatureTable(font, ligature) {
  let allCodePoints = font.codePoints

  let unicode = ligature.unicode

  let length =
    0 +
    2 + // LigGlyph
    2 + // CompCount
    2 * (unicode.length - 1)

  let buffer = new ByteBuffer(length)

  // LigGlyph
  let glyph = ligature.glyph

  buffer.writeUint16(glyph.id)

  // CompCount
  buffer.writeUint16(unicode.length)

  // Compound glyphs (excluding first as it’s already in the coverage table)
  for (let i = 1; i < unicode.length; i++) {
    glyph = allCodePoints[unicode[i]]
    buffer.writeUint16(glyph.id)
  }

  return buffer
}

function createLigatureSet(font, codePoint, ligatures) {
  let ligatureTables = []

  ligatures.forEach(function (ligature) {
    ligatureTables.push(createLigatureTable(font, ligature))
  })

  let tableLengths = ligatureTables
    .map(it => it.length)
    .reduce(function (result, count) {
      return result + count
    }, 0)

  let offset =
    0 +
    2 + // LigatureCount
    2 * ligatures.length

  let length = 0 + offset + tableLengths

  let buffer = new ByteBuffer(length)

  // LigatureCount
  buffer.writeUint16(ligatures.length)

  // Ligature offsets
  ligatureTables.forEach(function (table) {
    // The offset to the current set, from SubstFormat
    buffer.writeUint16(offset)
    offset += table.length
  })

  // Ligatures
  ligatureTables.forEach(function (table) {
    buffer.writeBytes(table.buffer)
  })

  return buffer
}

function createLigatureList(font, ligatureGroups) {
  let sets = []

  ligatureGroups.forEach(function (group) {
    let set = createLigatureSet(font, group.codePoint, group.ligatures)

    sets.push(set)
  })

  let setLengths = sets
    .map(it => it.length)
    .reduce(function (result, count) {
      return result + count
    }, 0)

  let coverage = createLigatureCoverage(font, ligatureGroups)

  let tableOffset =
    0 +
    2 + // Lookup type
    2 + // Lokup flag
    2 + // SubTableCount
    2 // SubTable[0] Offset

  let setOffset =
    0 +
    2 + // SubstFormat
    2 + // Coverage offset
    2 + // LigSetCount
    2 * sets.length // LigSet Offsets

  let coverageOffset = setOffset + setLengths

  let length = 0 + tableOffset + coverageOffset + coverage.length

  let buffer = new ByteBuffer(length)

  // Lookup type 4 – ligatures
  buffer.writeUint16(4)

  // Lookup flag – empty
  buffer.writeUint16(0)

  // Subtable count
  buffer.writeUint16(1)

  // Subtable[0] offset
  buffer.writeUint16(tableOffset)

  // SubstFormat
  buffer.writeUint16(1)

  // Coverage
  buffer.writeUint16(coverageOffset)

  // LigSetCount
  buffer.writeUint16(sets.length)

  sets.forEach(function (set) {
    // The offset to the current set, from SubstFormat
    buffer.writeUint16(setOffset)
    setOffset += set.length
  })

  sets.forEach(function (set) {
    buffer.writeBytes(set.buffer)
  })

  buffer.writeBytes(coverage.buffer)

  return buffer
}

// Add a lookup for each ligature
function createLookupList(font) {
  let ligatures = font.ligatures

  let groupedLigatures = {}

  // Group ligatures by first code point
  ligatures.forEach(function (ligature) {
    let first = ligature.unicode[0]

    if (!groupedLigatures[first]) {
      groupedLigatures[first] = []
    }
    groupedLigatures[first].push(ligature)
  })

  let ligatureGroups = []

  groupedLigatures.forEach(function (ligatures, codePoint) {
    codePoint = parseInt(codePoint, 10)
    // Order ligatures by length, descending
    // “Ligatures with more components must be stored ahead of those with fewer components in order to be found”
    // From: http://partners.adobe.com/public/developer/opentype/index_tag7.html#liga
    ligatures.sort(function (ligA, ligB) {
      return ligB.unicode.length - ligA.unicode.length
    })
    ligatureGroups.push({
      codePoint: codePoint,
      ligatures: ligatures,
      startGlyph: font.codePoints[codePoint]
    })
  })

  ligatureGroups.sort(function (a, b) {
    return a.startGlyph.id - b.startGlyph.id
  })

  let offset =
    0 +
    2 + // Lookup count
    2 // Lookup[0] offset

  let set = createLigatureList(font, ligatureGroups)

  let length = 0 + offset + set.length

  let buffer = new ByteBuffer(length)

  // Lookup count
  buffer.writeUint16(1)

  // Lookup[0] offset
  buffer.writeUint16(offset)

  // Lookup[0]
  buffer.writeBytes(set.buffer)

  return buffer
}

export default function createGSUB(font) {
  let scriptList = createScriptList()
  let featureList = createFeatureList()
  let lookupList = createLookupList(font)

  let lists = [scriptList, featureList, lookupList]

  let offset =
    0 +
    4 + // Version
    2 * lists.length // List offsets

  // Calculate offsets
  lists.forEach(function (list) {
    list._listOffset = offset
    offset += list.length
  })

  let length = offset
  let buffer = new ByteBuffer(length)

  // Version
  buffer.writeUint32(0x00010000)

  // Offsets
  lists.forEach(function (list) {
    buffer.writeUint16(list._listOffset)
  })

  // List contents
  lists.forEach(function (list) {
    buffer.writeBytes(list.buffer)
  })

  return buffer
}
